import warnings


class ExperimentalFeatureWarning(UserWarning):
    def __init__(self, feature_name: str):
        super().__init__(f"Experimental feature '{feature_name}' is subject to change without notice")

    def warn(self) -> None:
        warnings.warn(self, stacklevel=2)


# ExperimentalFlags to explicitly inform the user that these features are not stable!
class ExperimentalFlags:
    manual_rules_edit = ExperimentalFeatureWarning("enable_manual_edit")
    same_space_properties_only_export = ExperimentalFeatureWarning("same-space-properties-only")
    standardize_naming = ExperimentalFeatureWarning("standardize_naming")
    standardize_space_and_version = ExperimentalFeatureWarning("standardize_space_and_version")
    data_model_subsetting = ExperimentalFeatureWarning("data_model_subsetting")
    core_data_model_subsetting = ExperimentalFeatureWarning("core_data_model_subsetting")
    ontology_read = ExperimentalFeatureWarning("ontology_read")
    imf_read = ExperimentalFeatureWarning("imf_read")
    dexpi_read = ExperimentalFeatureWarning("dexpi_read")
    aml_read = ExperimentalFeatureWarning("aml_read")
    csv_read = ExperimentalFeatureWarning("csv_read")
    to_ontology = ExperimentalFeatureWarning("to_ontology")
    extension = ExperimentalFeatureWarning("extension")



from __future__ import annotations

import itertools
from datetime import datetime
from pathlib import Path
from types import GenericAlias
from typing import Any, ClassVar, Literal, cast, get_args

from openpyxl import Workbook
from openpyxl.cell import MergedCell
from openpyxl.styles import Alignment, Border, Font, PatternFill, Side
from openpyxl.worksheet.worksheet import Worksheet
from rdflib import Namespace

from cognite.neat._constants import COGNITE_CONCEPTS
from cognite.neat._rules._constants import get_internal_properties
from cognite.neat._rules._shared import VerifiedRules
from cognite.neat._rules.models import (
    SheetRow,
)
from cognite.neat._rules.models._base_rules import BaseMetadata, RoleTypes
from cognite.neat._rules.models.data_types import (
    _DATA_TYPE_BY_DMS_TYPE,
)
from cognite.neat._rules.models.dms._rules import DMSRules
from cognite.neat._rules.models.information._rules import InformationRules
from cognite.neat._utils.spreadsheet import (
    find_column_with_value,
    generate_data_validation,
)

from ._base import BaseExporter

MAX_COLUMN_WIDTH = 70.0


class ExcelExporter(BaseExporter[VerifiedRules, Workbook]):
    """Export rules to Excel.

    Args:
        styling: The styling to use for the Excel file. Defaults to "default". See below for details
            on the different styles.
        new_model_id: The new model ID to use for the exported spreadsheet. This is only applicable if the input
            rules have 'is_reference' set. If provided, the model ID will be used to automatically create the
            new metadata sheet in the Excel file. The model id is expected to be a tuple of (prefix, title)
            (space, external_id) for InformationRules and DMSRules respectively.

        sheet_prefix: The prefix to use for the sheet names in the Excel file. Defaults to an empty string.

    The following styles are available:

    - "none":    No styling is applied.
    - "minimal": Column widths are adjusted to fit the content, and the header row(s) is frozen.
    - "default": Minimal + headers are bold, increased size, and colored.
    - "maximal": Default + alternating row colors in the properties sheet for each class in addition to extra
                 blank rows between classes and borders
    """

    Style = Literal["none", "minimal", "default", "maximal"]
    DumpOptions = Literal["user", "last", "reference"]
    _helper_sheet_name: str = "_helper"
    _main_header_by_sheet_name: ClassVar[dict[str, str]] = {
        "Properties": "Definition of Properties",
        "Classes": "Definition of Classes",
        "Views": "Definition of Views",
        "Containers": "Definition of Containers",
        "Nodes": "Definition of Nodes",
        "Enum": "Definition of Enum Collections",
    }
    style_options = get_args(Style)
    dump_options = get_args(DumpOptions)

    def __init__(
        self,
        styling: Style = "default",
        new_model_id: tuple[str, str] | None = None,
        sheet_prefix: str | None = None,
        reference_rules_with_prefix: tuple[VerifiedRules, str] | None = None,
        add_empty_rows: bool = False,
        hide_internal_columns: bool = True,
        include_properties: Literal["same-space", "all"] = "all",
        add_drop_downs: bool = True,
    ):
        self.sheet_prefix = sheet_prefix or ""
        if styling not in self.style_options:
            raise ValueError(f"Invalid styling: {styling}. Valid options are {self.style_options}")
        self.styling = styling
        self._styling_level = self.style_options.index(styling)
        self.new_model_id = new_model_id
        self.reference_rules_with_prefix = reference_rules_with_prefix
        self.add_empty_rows = add_empty_rows
        self.hide_internal_columns = hide_internal_columns
        self.include_properties = include_properties
        self.add_drop_downs = add_drop_downs

    @property
    def description(self) -> str:
        return "Export verified model to Excel."

    def export_to_file(self, rules: VerifiedRules, filepath: Path) -> None:
        """Exports transformation rules to excel file."""
        data = self.export(rules)
        try:
            data.save(filepath)
        finally:
            data.close()
        return None

    def template(self, role: RoleTypes, filepath: Path | None = None) -> None | Workbook:
        """This method will create an spreadsheet template for data modeling depending on the role.

        Args:
            role: The role for which the template is created. Can be either "dms" or "information".
            filepath: The path to the file where the template will be saved.

        """
        workbook = Workbook()
        # Remove default sheet named "Sheet"
        workbook.remove(workbook["Sheet"])

        rules_model = DMSRules if role == RoleTypes.dms else InformationRules

        headers_by_sheet = rules_model.headers_by_sheet(by_alias=True)
        headers_by_sheet.pop("Metadata")

        self._write_metadata_sheet(
            workbook,
            cast(BaseMetadata, rules_model.model_fields["metadata"].annotation).default().model_dump(),
        )

        for sheet_name, headers in headers_by_sheet.items():
            if sheet_name in ("Metadata", "Prefixes", "Reference", "Last"):
                continue
            sheet = self._create_sheet_with_header(workbook, headers, sheet_name)
            self._style_sheet_header(sheet, headers)

        self._adjust_column_widths(workbook)
        self._hide_internal_columns(workbook)

        if role == RoleTypes.dms:
            self._add_dms_drop_downs(workbook)
        else:
            self._add_info_drop_downs(workbook)

        if filepath:
            try:
                workbook.save(filepath)
            finally:
                workbook.close()
            return None

        return workbook

    def export(self, rules: VerifiedRules) -> Workbook:
        workbook = Workbook()
        # Remove default sheet named "Sheet"
        workbook.remove(workbook["Sheet"])

        dumped_user_rules: dict[str, Any] = rules.dump(by_alias=True)

        self._write_metadata_sheet(workbook, dumped_user_rules["Metadata"], sheet_prefix=self.sheet_prefix)
        self._write_sheets(workbook, dumped_user_rules, rules, sheet_prefix=self.sheet_prefix)
        if self.reference_rules_with_prefix:
            reference_rules, prefix = self.reference_rules_with_prefix
            dumped_reference_rules = reference_rules.dump(entities_exclude_defaults=False, by_alias=True)
            self._write_sheets(workbook, dumped_reference_rules, reference_rules, sheet_prefix=prefix)
            self._write_metadata_sheet(workbook, dumped_reference_rules["Metadata"], sheet_prefix=prefix)

        if isinstance(rules, InformationRules) and rules.prefixes:
            self._write_prefixes_sheet(workbook, rules.prefixes)

        if self._styling_level > 0:
            self._adjust_column_widths(workbook)

        if self.hide_internal_columns:
            self._hide_internal_columns(workbook)

        # Only add drop downs if the rules are DMSRules
        if self.add_drop_downs and isinstance(rules, DMSRules):
            self._add_dms_drop_downs(workbook)
        elif self.add_drop_downs and isinstance(rules, InformationRules):
            self._add_info_drop_downs(workbook)

        return workbook

    def _hide_internal_columns(self, workbook: Workbook) -> None:
        """Hides internal columns in workbook sheets.

        Args:
            workbook: Workbook representation of the Excel file.

        """
        for sheet in workbook.sheetnames:
            if sheet.lower() == "metadata":
                continue
            ws = workbook[sheet]
            for col in get_internal_properties():
                column_letter = find_column_with_value(ws, col)
                if column_letter:
                    ws.column_dimensions[column_letter].hidden = True

    def _add_info_drop_downs(self, workbook: Workbook, no_rows: int = 100) -> None:
        """Adds drop down menus to specific columns for fast and accurate data entry
        in the Information rules.

        Args:
            workbook: Workbook representation of the Excel file.
            no_rows: number of rows to add drop down menus. Defaults to 100*100.

        !!! note "Why no_rows=100?"
            Maximum number of views per data model is 100, thus this value is set accordingly

        !!! note "Why defining individual data validation per desired column?
            This is due to the internal working of openpyxl. Adding same validation to
            different column leads to unexpected behavior when the openpyxl workbook is exported
            as and Excel file. Probably, the validation is not copied to the new column,
            but instead reference to the data validation object is added.
        """
        self._make_helper_info_sheet(workbook, no_rows)

        # We need create individual data validation and cannot re-use the same one due
        # the internals of openpyxl
        dv_classes = generate_data_validation(self._helper_sheet_name, "A", no_header_rows=0, no_rows=no_rows)
        dv_value_types = generate_data_validation(self._helper_sheet_name, "B", no_header_rows=0, no_rows=no_rows)
        dv_implements = generate_data_validation(
            self._helper_sheet_name,
            "C",
            no_header_rows=0,
            no_rows=no_rows + len(COGNITE_CONCEPTS),
        )

        workbook["Properties"].add_data_validation(dv_classes)
        workbook["Properties"].add_data_validation(dv_value_types)
        workbook["Classes"].add_data_validation(dv_implements)

        # we multiply no_rows with 100 since a view can have max 100 properties per view
        if column := find_column_with_value(workbook["Properties"], "Class"):
            dv_classes.add(f"{column}{3}:{column}{no_rows * 100}")

        if column := find_column_with_value(workbook["Properties"], "Value Type"):
            dv_value_types.add(f"{column}{3}:{column}{no_rows * 100}")

        if column := find_column_with_value(workbook["Classes"], "Implements"):
            dv_implements.add(f"{column}{3}:{column}{no_rows}")

    def _make_helper_info_sheet(self, workbook: Workbook, no_rows: int) -> None:
        """This helper Information sheet is used as source of data for drop down menus creation"""
        workbook.create_sheet(title=self._helper_sheet_name)

        for dtype_counter, dtype in enumerate(_DATA_TYPE_BY_DMS_TYPE.values()):
            # skip types which require special handling or are surpassed by CDM
            if dtype.xsd in ["enum", "timeseries", "sequence", "file", "json"]:
                continue
            workbook[self._helper_sheet_name].cell(row=dtype_counter + 1, column=2, value=dtype.xsd)

        # Add Cognite Core Data Views:
        for concept_counter, concept in enumerate(COGNITE_CONCEPTS):
            workbook[self._helper_sheet_name].cell(
                row=concept_counter + 1,
                column=3,
                value=f"cdf_cdm:{concept}(version=v1)",
            )

        for i in range(no_rows):
            workbook[self._helper_sheet_name].cell(
                row=i + 1,
                column=1,
                value=f'=IF(ISBLANK(Classes!A{i + 3}), "", Classes!A{i + 3})',
            )
            workbook[self._helper_sheet_name].cell(
                row=dtype_counter + i + 2,
                column=2,
                value=f'=IF(ISBLANK(Classes!A{i + 3}), "", Classes!A{i + 3})',
            )
            workbook[self._helper_sheet_name].cell(
                row=concept_counter + i + 2,
                column=3,
                value=f'=IF(ISBLANK(Classes!A{i + 3}), "", Classes!A{i + 3})',
            )

        workbook[self._helper_sheet_name].sheet_state = "hidden"

    def _add_dms_drop_downs(self, workbook: Workbook, no_rows: int = 100) -> None:
        """Adds drop down menus to specific columns for fast and accurate data entry
        in the DMS rules.

        Args:
            workbook: Workbook representation of the Excel file.
            no_rows: number of rows to add drop down menus. Defaults to 100*100.

        !!! note "Why no_rows=100?"
            Maximum number of views per data model is 100, thus this value is set accordingly

        !!! note "Why defining individual data validation per desired column?
            This is due to the internal working of openpyxl. Adding same validation to
            different column leads to unexpected behavior when the openpyxl workbook is exported
            as and Excel file. Probably, the validation is not copied to the new column,
            but instead reference to the data validation object is added.
        """

        self._make_helper_dms_sheet(workbook, no_rows)

        # We need create individual data validation and cannot re-use the same one due
        # the internals of openpyxl
        dv_views = generate_data_validation(self._helper_sheet_name, "A", no_header_rows=0, no_rows=no_rows)
        dv_containers = generate_data_validation(self._helper_sheet_name, "B", no_header_rows=0, no_rows=no_rows)
        dv_value_types = generate_data_validation(self._helper_sheet_name, "C", no_header_rows=0, no_rows=no_rows)
        dv_implements = generate_data_validation(
            self._helper_sheet_name,
            "F",
            no_header_rows=0,
            no_rows=no_rows + len(COGNITE_CONCEPTS),
        )

        dv_immutable = generate_data_validation(self._helper_sheet_name, "D", no_header_rows=0, no_rows=3)
        dv_in_model = generate_data_validation(self._helper_sheet_name, "D", no_header_rows=0, no_rows=3)
        dv_used_for = generate_data_validation(self._helper_sheet_name, "E", no_header_rows=0, no_rows=3)

        workbook["Properties"].add_data_validation(dv_views)
        workbook["Properties"].add_data_validation(dv_containers)
        workbook["Properties"].add_data_validation(dv_value_types)
        workbook["Properties"].add_data_validation(dv_immutable)
        workbook["Views"].add_data_validation(dv_in_model)
        workbook["Views"].add_data_validation(dv_implements)
        workbook["Containers"].add_data_validation(dv_used_for)

        # we multiply no_rows with 100 since a view can have max 100 properties per view
        if column := find_column_with_value(workbook["Properties"], "View"):
            dv_views.add(f"{column}{3}:{column}{no_rows * 100}")

        if column := find_column_with_value(workbook["Properties"], "Container"):
            dv_containers.add(f"{column}{3}:{column}{no_rows * 100}")

        if column := find_column_with_value(workbook["Properties"], "Value Type"):
            dv_value_types.add(f"{column}{3}:{column}{no_rows * 100}")

        if column := find_column_with_value(workbook["Properties"], "Immutable"):
            dv_immutable.add(f"{column}{3}:{column}{no_rows * 100}")

        if column := find_column_with_value(workbook["Views"], "In Model"):
            dv_in_model.add(f"{column}{3}:{column}{no_rows}")

        if column := find_column_with_value(workbook["Views"], "Implements"):
            dv_implements.add(f"{column}{3}:{column}{no_rows}")

        if column := find_column_with_value(workbook["Containers"], "Used For"):
            dv_used_for.add(f"{column}{3}:{column}{no_rows}")

    def _make_helper_dms_sheet(self, workbook: Workbook, no_rows: int) -> None:
        """This helper DMS sheet is used as source of data for drop down menus creation"""
        workbook.create_sheet(title=self._helper_sheet_name)

        for dtype_counter, dtype in enumerate(_DATA_TYPE_BY_DMS_TYPE):
            if dtype in ["enum", "timeseries", "sequence", "file", "json"]:
                continue
            workbook[self._helper_sheet_name].cell(row=dtype_counter + 1, column=3, value=dtype)

        # Add Cognite Core Data Views:
        for concept_counter, concept in enumerate(COGNITE_CONCEPTS):
            workbook[self._helper_sheet_name].cell(
                row=concept_counter + 1,
                column=6,
                value=f"cdf_cdm:{concept}(version=v1)",
            )

        for i in range(no_rows):
            workbook[self._helper_sheet_name].cell(
                row=i + 1,
                column=1,
                value=f'=IF(ISBLANK(Views!A{i + 3}), "", Views!A{i + 3})',
            )
            workbook[self._helper_sheet_name].cell(
                row=i + 1,
                column=2,
                value=f'=IF(ISBLANK(Containers!A{i + 3}), "", Containers!A{i + 3})',
            )
            workbook[self._helper_sheet_name].cell(
                row=dtype_counter + i + 2,
                column=3,
                value=f'=IF(ISBLANK(Views!A{i + 3}), "", Views!A{i + 3})',
            )
            workbook[self._helper_sheet_name].cell(
                row=concept_counter + i + 2,
                column=6,
                value=f'=IF(ISBLANK(Views!A{i + 3}), "", Views!A{i + 3})',
            )

        for i, value in enumerate([True, False, ""]):
            workbook[self._helper_sheet_name].cell(row=i + 1, column=4, value=cast(bool | str, value))

        for i, value in enumerate(["node", "edge", "all"]):
            workbook[self._helper_sheet_name].cell(row=i + 1, column=5, value=value)

        workbook[self._helper_sheet_name].sheet_state = "hidden"

    def _create_sheet_with_header(
        self,
        workbook: Workbook,
        headers: list[str],
        sheet_name: str,
        sheet_prefix: str = "",
    ) -> Worksheet:
        """Creates an empty sheet with the given headers.

        Args:
            workbook: The workbook to add the sheet to.
            headers: The headers to add to the sheet.
            sheet_name: The name of the sheet.
            sheet_prefix: The prefix to add to the sheet name, if any.
        """

        sheet = workbook.create_sheet(f"{sheet_prefix}{sheet_name}")
        main_header = self._main_header_by_sheet_name[sheet_name]
        sheet.append([main_header] + [""] * (len(headers) - 1))

        if headers[0] == "Neat ID":
            # Move the Neat ID to the end of the columns
            headers = headers[1:] + ["Neat ID"]

        # Append the headers to the sheet
        sheet.append(headers)

        return sheet

    def _style_sheet_header(self, sheet: Worksheet, headers: list[str]) -> None:
        """Styles the sheet with the given headers.

        Args:
            sheet: The sheet to style.
            headers: The headers to style.
        """
        if self._styling_level > 0:
            # This freezes all rows above the given row
            sheet.freeze_panes = sheet["A3"]

            sheet["A1"].alignment = Alignment(horizontal="left")

        if self._styling_level > 1:
            # Make the header row bold, larger, and colored
            for cell, *_ in sheet.iter_cols(min_row=1, max_row=1, min_col=1, max_col=len(headers)):
                cell.font = Font(bold=True, size=20)
                cell.fill = PatternFill(fgColor="FFC000", patternType="solid")
            for cell in sheet["2"]:
                cell.font = Font(bold=True, size=14)

    def _write_sheets(
        self,
        workbook: Workbook,
        dumped_rules: dict[str, Any],
        rules: VerifiedRules,
        sheet_prefix: str = "",
    ) -> None:
        for sheet_name, headers in rules.headers_by_sheet(by_alias=True).items():
            if sheet_name in ("Metadata", "Prefixes", "Reference", "Last"):
                continue

            sheet = self._create_sheet_with_header(workbook, headers, sheet_name, sheet_prefix)

            fill_colors = itertools.cycle(["CADCFC", "FFFFFF"])
            fill_color = next(fill_colors)
            last_class: str | None = None
            item: dict[str, Any]
            for item in dumped_rules.get(sheet_name) or []:
                if "Neat ID" in item:
                    # Move the Neat ID to the end of the columns
                    item["Neat ID"] = item.pop("Neat ID")
                row = list(item.values())
                class_ = row[0]

                is_properties = sheet_name == "Properties"
                is_new_class = class_ != last_class and last_class is not None
                if self._styling_level > 2 and is_new_class and is_properties:
                    if self.add_empty_rows:
                        sheet.append([""] * len(headers))
                    for cell in sheet[sheet.max_row]:
                        cell.fill = PatternFill(fgColor=fill_color, patternType="solid")
                        side = Side(style="thin", color="000000")
                        cell.border = Border(left=side, right=side, top=side, bottom=side)
                    fill_color = next(fill_colors)

                if is_properties and self.include_properties == "same-space":
                    space = class_.split(":")[0] if ":" in class_ else rules.metadata.space
                    if space != rules.metadata.space:
                        continue

                sheet.append(row)
                if self._styling_level > 2 and is_properties:
                    for cell in sheet[sheet.max_row]:
                        cell.fill = PatternFill(fgColor=fill_color, patternType="solid")
                        side = Side(style="thin", color="000000")
                        cell.border = Border(left=side, right=side, top=side, bottom=side)
                last_class = class_

            self._style_sheet_header(sheet, headers)

    def _write_metadata_sheet(self, workbook: Workbook, metadata: dict[str, Any], sheet_prefix: str = "") -> None:
        # Excel does not support timezone in datetime strings
        if isinstance(metadata.get("created"), datetime):
            metadata["created"] = metadata["created"].replace(tzinfo=None)
        if isinstance(metadata.get("updated"), datetime):
            metadata["updated"] = metadata["updated"].replace(tzinfo=None)

        metadata_sheet = workbook.create_sheet(f"{sheet_prefix}Metadata")
        for key, value in metadata.items():
            metadata_sheet.append([key, value])

        if self._styling_level > 1:
            for cell in metadata_sheet["A"]:
                cell.font = Font(bold=True, size=12)

    def _write_prefixes_sheet(self, workbook: Workbook, prefixes: dict[str, Namespace]) -> None:
        metadata_sheet = workbook.create_sheet("Prefixes")
        metadata_sheet.append(["Prefix", "Namespace"])
        for key, value in prefixes.items():
            metadata_sheet.append([key, value])

        if self._styling_level > 1:
            for cell in metadata_sheet["A"]:
                cell.font = Font(bold=True, size=12)

    @classmethod
    def _get_item_class(cls, annotation: GenericAlias) -> type[SheetRow]:
        if not isinstance(annotation, GenericAlias):
            raise ValueError(f"Expected annotation to be a GenericAlias, but got {type(annotation)}")
        args = get_args(annotation)
        if len(args) != 1:
            raise ValueError(f"Expected annotation to have exactly one argument, but got {len(args)}")
        arg = args[0]
        if not issubclass(arg, SheetRow):
            raise ValueError(f"Expected annotation to have a BaseModel argument, but got {type(arg)}")
        return arg

    @classmethod
    def _adjust_column_widths(cls, workbook: Workbook) -> None:
        for sheet_ in workbook:
            sheet = cast(Worksheet, sheet_)
            for column_cells in sheet.columns:
                try:
                    max_length = max(len(str(cell.value)) for cell in column_cells if cell.value is not None)
                except ValueError:
                    max_length = 0

                selected_column = column_cells[0]
                if isinstance(selected_column, MergedCell):
                    selected_column = column_cells[1]

                current = sheet.column_dimensions[selected_column.column_letter].width or (max_length + 0.5)  # type: ignore[union-attr]
                sheet.column_dimensions[selected_column.column_letter].width = min(  # type: ignore[union-attr]
                    max(current, max_length + 0.5), MAX_COLUMN_WIDTH
                )
        return None



"""This module contains the definition of `TransformationRules` pydantic model and all
its sub-models and validators.
"""

import math
import sys
import types
from abc import ABC, abstractmethod
from collections.abc import Callable, Hashable, Iterator, MutableSequence, Sequence
from datetime import datetime
from typing import Annotated, Any, ClassVar, Literal, SupportsIndex, TypeVar, get_args, get_origin, overload

import pandas as pd
from cognite.client import data_modeling as dm
from pydantic import (
    BaseModel,
    BeforeValidator,
    ConfigDict,
    Field,
    GetCoreSchemaHandler,
    PlainSerializer,
    field_validator,
    model_serializer,
)
from pydantic.main import IncEx
from pydantic_core import core_schema
from rdflib import Namespace, URIRef

from cognite.neat._constants import DEFAULT_NAMESPACE
from cognite.neat._rules.models._types import (
    ContainerEntityType,
    DataModelExternalIdType,
    DmsPropertyType,
    SpaceType,
    StrListType,
    URIRefType,
    VersionType,
    ViewEntityType,
)
from cognite.neat._rules.models.data_types import DataType
from cognite.neat._rules.models.entities import EdgeEntity, ReverseConnectionEntity, ViewEntity
from cognite.neat._utils.rdf_ import uri_display_name

if sys.version_info >= (3, 11):
    from enum import StrEnum
else:
    from backports.strenum import StrEnum


METADATA_VALUE_MAX_LENGTH = 5120


def _get_required_fields(model: type[BaseModel], use_alias: bool = False) -> set[str]:
    """Get required fields from a pydantic model.

    Parameters
    ----------
    model : type[BaseModel]
        Pydantic data model
    use_alias : bool, optional
        Whether to return field alias name, by default False

    Returns
    -------
    list[str]
        List of required fields
    """
    required_fields = set()
    for name, field in model.model_fields.items():
        if not field.is_required():
            continue

        alias = getattr(field, "alias", None)
        if use_alias and alias:
            required_fields.add(alias)
        else:
            required_fields.add(name)
    return required_fields


class SchemaCompleteness(StrEnum):
    complete = "complete"
    partial = "partial"
    extended = "extended"


class ExtensionCategory(StrEnum):
    addition = "addition"
    reshape = "reshape"
    rebuild = "rebuild"


class DataModelType(StrEnum):
    solution = "solution"
    enterprise = "enterprise"


class DataModelAspect(StrEnum):
    conceptual = "conceptual"
    logical = "logical"
    physical = "physical"


class RoleTypes(StrEnum):
    domain_expert = "domain expert"
    information = "information architect"
    dms = "DMS Architect"


class SchemaModel(BaseModel):
    model_config: ClassVar[ConfigDict] = ConfigDict(
        populate_by_name=True,
        str_strip_whitespace=True,
        arbitrary_types_allowed=True,
        strict=False,
        extra="ignore",
        use_enum_values=True,
    )

    @classmethod
    def mandatory_fields(cls: Any, use_alias: bool = False) -> set[str]:
        """Returns a set of mandatory fields for the model."""
        return _get_required_fields(cls, use_alias)

    @field_validator("*", mode="before")
    def strip_string(cls, value: Any) -> Any:
        if isinstance(value, str):
            return value.strip()
        elif isinstance(value, list):
            return [entry.strip() if isinstance(entry, str) else entry for entry in value]
        return value


class BaseMetadata(SchemaModel):
    """
    Metadata model for data model
    """

    role: ClassVar[RoleTypes] = Field(description="Role of the person creating the data model")
    aspect: ClassVar[DataModelAspect] = Field(description="Aspect of the data model")
    space: SpaceType = Field(description="The space where the data model is defined")
    external_id: DataModelExternalIdType = Field(
        alias="externalId", description="External identifier for the data model"
    )
    version: VersionType = Field(description="Version of the data model")

    name: str | None = Field(
        None,
        description="Human readable name of the data model",
        min_length=1,
        max_length=255,
    )

    description: str | None = Field(
        None, min_length=1, max_length=1024, description="Short description of the data model"
    )

    creator: StrListType = Field(
        description=(
            "List of contributors (comma seperated) to the data model creation, "
            "typically information architects are considered as contributors."
        ),
    )

    created: datetime = Field(
        description="Date of the data model creation",
    )

    updated: datetime = Field(
        description="Date of the data model update",
    )

    source_id: URIRefType | None = Field(
        None,
        description="Id of source that produced this rules",
        alias="sourceId",
    )

    @field_validator("*", mode="before")
    def strip_string(cls, value: Any) -> Any:
        if isinstance(value, str):
            return value.strip()
        return value

    @field_validator("description", mode="before")
    def nan_as_none(cls: Any, value: Any) -> Any:
        if isinstance(value, float) and math.isnan(value):
            return None
        return value

    def to_pandas(self) -> pd.Series:
        """Converts Metadata to pandas Series."""
        return pd.Series(self.model_dump())

    def _repr_html_(self) -> str:
        """Returns HTML representation of Metadata."""
        return self.to_pandas().to_frame("value")._repr_html_()  # type: ignore[operator]

    @classmethod
    def mandatory_fields(cls: Any, use_alias: bool = False) -> set[str]:
        """Returns a set of mandatory fields for the model."""
        return _get_required_fields(cls, use_alias)

    @model_serializer(mode="wrap")
    def include_role(self, serializer: Callable) -> dict:
        return {"role": self.role.value, **serializer(self)}

    @property
    def prefix(self) -> str:
        return self.space

    def get_prefix(self) -> str:
        return self.prefix

    @property
    def identifier(self) -> URIRef:
        """Globally unique identifier for the data model.

        !!! note
            Unlike namespace, the identifier does not end with "/" or "#".

        """
        return DEFAULT_NAMESPACE[f"data-model/verified/{self.aspect}/{self.space}/{self.external_id}/{self.version}"]

    @property
    def namespace(self) -> Namespace:
        """Namespace for the data model used for the entities in the data model."""
        return Namespace(f"{self.identifier}/")

    def as_data_model_id(self) -> dm.DataModelId:
        return dm.DataModelId(space=self.space, external_id=self.external_id, version=self.version)

    def as_identifier(self) -> str:
        return repr(self.as_data_model_id())

    @classmethod
    def default(cls) -> "BaseMetadata":
        """Returns a default instance of the metadata model."""
        now = datetime.now()
        return cls(
            space="pleaseUpdateMe",
            external_id="PleaseUpdateMe",
            version="v1",
            name="Please Update Me",
            description="Please Update Me",
            creator=["NEAT"],
            created=now,
            updated=now,
        )


class BaseRules(SchemaModel, ABC):
    """
    Rules is a core concept in `neat`. This represents fusion of data model
    definitions and (optionally) the transformation rules used to transform the data/graph
    from the source representation to the target representation defined by the data model.
    The rules are defined in an Excel sheet and then parsed into a `Rules` object. The
    `Rules` object is then used to generate data model and the `RDF` graph made of data
    model instances.

    Args:
        metadata: Data model metadata
    """

    metadata: BaseMetadata

    @classmethod
    def headers_by_sheet(cls, by_alias: bool = False) -> dict[str, list[str]]:
        """Returns a list of headers for the model, typically used by ExcelExporter"""
        headers_by_sheet: dict[str, list[str]] = {}
        for field_name, field in cls.model_fields.items():
            if field_name in ["validators_to_skip", "post_validate"]:
                continue
            sheet_name = (field.alias or field_name) if by_alias else field_name
            annotation = field.annotation

            if isinstance(annotation, types.UnionType):
                annotation = annotation.__args__[0]

            try:
                if isinstance(annotation, types.GenericAlias) and get_origin(annotation).__name__ == SheetList.__name__:
                    # We know that this is a SheetList, so we can safely access the annotation
                    # which is the concrete type of the SheetEntity.
                    model_fields = get_args(annotation)[0].model_fields  # type: ignore[union-attr]
                elif isinstance(annotation, type) and issubclass(annotation, BaseModel):
                    model_fields = annotation.model_fields
                else:
                    model_fields = {}
            except TypeError:
                # Python 3.10 raises TypeError: issubclass() arg 1 must be a class
                # when calling issubclass(annotation, SheetList) with the dict annotation
                model_fields = {}
            headers_by_sheet[sheet_name] = [
                (field.alias or field_name) if by_alias else field_name
                for field_name, field in model_fields.items()
                if field_name != "validators_to_skip" and not field.exclude
            ]
        return headers_by_sheet

    @property
    def display_name(self) -> str:
        return uri_display_name(self.metadata.identifier)

    def dump(
        self,
        entities_exclude_defaults: bool = True,
        sort: bool = False,
        mode: Literal["python", "json"] = "python",
        by_alias: bool = False,
        exclude: IncEx | None = None,
        exclude_none: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
    ) -> dict[str, Any]:
        """Dump the model to a dictionary.

        This is used in the Exporters to dump rules in the required format.

        Args:
            entities_exclude_defaults: Whether to exclude default prefix (and version) for entities.
                For example, given a class that is dumped as 'my_prefix:MyClass', if the prefix for the rules
                set in metadata.prefix = 'my_prefix', then this class will be dumped as 'MyClass' when this flag is set.
                Defaults to True.
            sort: Whether to sort the entities in the output.
            mode: The mode in which `to_python` should run.
                If mode is 'json', the output will only contain JSON serializable types.
                If mode is 'python', the output may contain non-JSON-serializable Python objects.
            by_alias: Whether to use the field's alias in the dictionary key if defined.
            exclude: A set of fields to exclude from the output.
            exclude_none: Whether to exclude fields that have a value of `None`.
            exclude_unset: Whether to exclude fields that have not been explicitly set.
            exclude_defaults: Whether to exclude fields that are set to their default value.
        """
        if sort:
            for field_name in self.model_fields.keys():
                value = getattr(self, field_name)
                # Ensure deterministic order of properties, classes, views, and so on
                if isinstance(value, SheetList):
                    value.sort(key=lambda x: x._identifier())

        context: dict[str, Any] = {}
        if entities_exclude_defaults:
            context["metadata"] = self.metadata

        exclude_input: IncEx | None = exclude

        return self.model_dump(
            mode=mode,
            by_alias=by_alias,
            exclude=exclude_input,
            exclude_none=exclude_none,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            context=context,
        )


class SheetRow(SchemaModel):
    neatId: URIRefType | None = Field(
        alias="Neat ID",
        description="Globally unique identifier for the property",
        default=None,
    )

    @abstractmethod
    def _identifier(self) -> tuple[Hashable, ...]:
        raise NotImplementedError()

    def __repr__(self) -> str:
        # Simplified representation of the object for debugging
        return f"{self.__class__.__name__}({self._identifier()})"


T_SheetRow = TypeVar("T_SheetRow", bound=SheetRow)


class SheetList(list, MutableSequence[T_SheetRow]):
    @classmethod
    def __get_pydantic_core_schema__(cls, source: Any, handler: GetCoreSchemaHandler) -> core_schema.CoreSchema:
        if args := get_args(source):
            item_type = args[0]
        else:
            # Someone use SheetList without specifying the type
            raise TypeError("SheetList must be used with a type argument, e.g., SheetList[InformationProperty]")

        instance_schema = core_schema.is_instance_schema(cls)
        sequence_row_schema = handler.generate_schema(Sequence[item_type])  # type: ignore[valid-type]

        non_instance_schema = core_schema.no_info_after_validator_function(SheetList, sequence_row_schema)
        return core_schema.union_schema([instance_schema, non_instance_schema])

    def to_pandas(self, drop_na_columns: bool = True, include: list[str] | None = None) -> pd.DataFrame:
        """Converts ResourceDict to pandas DataFrame."""
        df = pd.DataFrame([entity.model_dump() for entity in self])
        if drop_na_columns:
            df = df.dropna(axis=1, how="all")
        if include is not None:
            df = df[include]
        return df

    def _repr_html_(self) -> str:
        """Returns HTML representation of ResourceDict."""
        df = self.to_pandas(drop_na_columns=True)
        if "neatId" in df.columns:
            df = df.drop(columns=["neatId"])
        return df._repr_html_()  # type: ignore[operator]

    # Implemented to get correct type hints
    def __iter__(self) -> Iterator[T_SheetRow]:
        return super().__iter__()

    @overload
    def __getitem__(self, index: SupportsIndex) -> T_SheetRow: ...

    @overload
    def __getitem__(self, index: slice) -> "SheetList[T_SheetRow]": ...

    def __getitem__(self, index: SupportsIndex | slice, /) -> "T_SheetRow | SheetList[T_SheetRow]":
        if isinstance(index, slice):
            return SheetList[T_SheetRow](super().__getitem__(index))
        return super().__getitem__(index)


ExtensionCategoryType = Annotated[
    ExtensionCategory,
    PlainSerializer(
        lambda v: v.value if isinstance(v, ExtensionCategory) else v,
        return_type=str,
        when_used="unless-none",
    ),
    BeforeValidator(lambda v: ExtensionCategory(v) if isinstance(v, str) else v),
]


# Immutable such that this can be used as a key in a dictionary
class ContainerProperty(BaseModel, frozen=True):
    container: ContainerEntityType
    property_: DmsPropertyType


class ContainerDestinationProperty(ContainerProperty, frozen=True):
    value_type: DataType | ViewEntity
    connection: Literal["direct"] | ReverseConnectionEntity | EdgeEntity | None = None


class ViewRef(BaseModel, frozen=True):
    view: ViewEntityType


class ViewProperty(ViewRef, frozen=True):
    property_: DmsPropertyType



from pathlib import Path
from typing import Any, Literal

from cognite.client.data_classes.data_modeling import DataModelIdentifier

from cognite.neat._experimental._flags import ExperimentalFlags
from cognite.neat._issues import IssueList, catch_issues
from cognite.neat._rules._shared import ReadRules
from cognite.neat._rules.exporters import ExcelExporter
from cognite.neat._rules.importers import ExcelImporter
from cognite.neat._rules.models import InformationInputRules
from cognite.neat._rules.models._base_rules import RoleTypes
from cognite.neat._rules.models.dms import DMSValidation
from cognite.neat._rules.transformers import (
    AddCogniteProperties,
    IncludeReferenced,
    ToDataProductModel,
    ToEnterpriseModel,
    VerifiedRulesTransformer,
)
from cognite.neat._utils.reader import NeatReader, PathReader

from ._state import SessionState
from .exceptions import NeatSessionError, session_class_wrapper


@session_class_wrapper
class TemplateAPI:
    """
    Create a template for a new data model.
    """

    def __init__(self, state: SessionState):
        self._state = state

    def enterprise_model(
        self,
        data_model_id: DataModelIdentifier,
        org_name: str = "CopyOf",
        dummy_property: str = "GUID",
    ) -> IssueList:
        """Creates a template for an enterprise model based on the current data model in the session.
        An enterprise data model is a model that is used for read and write of instances. In addition,
        it is governed by the organization.
        The basis for an enterprise data model should be a Cognite Data Model.

        Args:
            data_model_id: The enterprise data model id that is being created
            org_name: Organization name to use for the views in the enterprise data model.
            dummy_property: The dummy property to use as placeholder for the views in the new data model.

        What does this function do?
            1. It creates a new view for each view in the current data model that implements the view it is based on.
            2. If dummy_property is set, it will create a container with one property for each view and connect the
               view to the container.
            3. It will repeat all connection properties in the new views and update the ValueTypes to match the new
               views.

        !!! note "Enterprise Data Model Creation"

            Always create an enterprise data model from a Cognite Data Model as this will
            assure all the Cognite Data Fusion applications to run smoothly, such as
                - Search
                - Atlas AI
                - Infield
                - Canvas
                - Maintain
                - Charts

        """
        last_rules = self._state.rule_store.last_verified_rules
        issues = self._state.rule_transform(
            ToEnterpriseModel(
                new_model_id=data_model_id,
                org_name=org_name,
                dummy_property=dummy_property,
                move_connections=True,
            )
        )
        if last_rules and not issues.has_errors:
            self._state.last_reference = last_rules
        return issues

    def data_product_model(
        self,
        data_model_id: DataModelIdentifier,
        include: Literal["same-space", "all"] = "same-space",
    ) -> IssueList:
        """Creates a template for a data product model based on the current data model in the session.
        A data product model is only used for reading of instances.
        It can be based on any data model, but typically it is based on an enterprise or solution data model.

        A data product model is a data model that ONLY maps to containers and do not use implements. This is
        typically used for defining the data in a data product.

        What does this function do?
        1. It creates a new view for each view in the current data model. The new views uses the same filter
           as the view it is based on.
        2. It will repeat all connection properties in the new views and update the ValueTypes to match the new views.

        Args:
            data_model_id: The data product data model id that is being created.
            include: The views to include in the data product data model. Can be either "same-space" or "all".
                If you set same-space, only the properties of the views in the same space as the data model
                will be included.
        """
        last_rules = self._state.rule_store.last_verified_rules
        view_ids, container_ids = DMSValidation(
            self._state.rule_store.last_verified_dms_rules
        ).imported_views_and_containers_ids()
        transformers: list[VerifiedRulesTransformer] = []
        client = self._state.client
        if (view_ids or container_ids) and client is None:
            raise NeatSessionError(
                "No client provided. You are referencing unknown views and containers in your data model, "
                "NEAT needs a client to lookup the definitions. "
                "Please set the client in the session, NeatSession(client=client)."
            )
        elif (view_ids or container_ids) and client:
            transformers.append(IncludeReferenced(client, include_properties=True))

        transformers.append(ToDataProductModel(new_model_id=data_model_id, include=include))

        issues = self._state.rule_transform(*transformers)
        if last_rules and not issues.has_errors:
            self._state.last_reference = last_rules
        return issues

    def conceptual_model(self, io: Any) -> None:
        """This method will create a template for a conceptual data modeling

        Args:
            io: file path to the Excel sheet

        """
        reader = NeatReader.create(io)
        path = reader.materialize_path()

        ExcelExporter().template(RoleTypes.information, path)

        return None

    def expand(self, io: Any, output: str | Path | None = None, dummy_property: str = "GUID") -> IssueList:
        """Creates a template for an extension of a Cognite model by expanding properties from CDM.

        The input is a spreadsheet of a conceptual model in which the concepts are defined
        and marked with the Cognite concept they are extending. For example, if you have a pump
        in the Classes sheet you will see
        ```
        Class: Pump
        Implements: cdf_cdm:CogniteAsset(version=v1)
        ```
        The output will be a spreadsheet in which all the properties from the Cognite concept model
        is added to the spreadsheet. In the example above, the pump concept will have all
        the properties it inherits from the CogniteAsset concept added to the Properties spreadsheet.


        Args:
            io: The input spreadsheet.
            output: The output spreadsheet. If None, the output will be the same
                as the input with `_extension` added to the name.
            dummy_property: The dummy property to use as placeholder for user-defined properties
                for each user-defined concept, and to alleviate need for usage of filters in
                physical data model. When converting a data model, it is recommended to have at least
                one property for each concept. This ensures that you follow that recommendation.
        """
        ExperimentalFlags.extension.warn()
        reader = NeatReader.create(io)
        path = reader.materialize_path()
        if output is None:
            if isinstance(reader, PathReader):
                output_path = path.with_name(f"{path.stem}_expand{path.suffix}")
            else:
                # The source is not a file, for example, a URL or a stream.
                output_path = Path.cwd() / f"{path.stem}_expand{path.suffix}"
        else:
            output_path = Path(output)

        with catch_issues() as issues:
            read: ReadRules[InformationInputRules] = ExcelImporter(path).to_rules()
            if read.rules is not None:
                # If rules are None there will be issues that are already caught.
                if not isinstance(read.rules, InformationInputRules):
                    raise NeatSessionError(f"The input {reader.name} must contain an InformationInputRules object. ")
                if self._state.client is None:
                    raise NeatSessionError("Client must be set in the session to run the extension.")
                modified = AddCogniteProperties(self._state.client, dummy_property).transform(read)
                if modified.rules is not None:
                    # If rules are None there will be issues that are already caught.
                    info = modified.rules.as_verified_rules()

                    ExcelExporter(styling="maximal").export_to_file(info, output_path)
        issues.action = "Created extension template"

        # Adding issues to the state in the rule store
        if issues:
            self._state.rule_store._last_issues = issues
        return issues



from rdflib import Dataset, URIRef

from cognite.neat._utils.rdf_ import remove_instance_ids_in_batch

from ._base import BaseQuery
from ._select import SelectQueries


class UpdateQueries(BaseQuery):
    """This class holds a set of SPARQL queries which are updating triples in the knowledge graph.
    The update queries are executed against update endpoint, and typically start with UPDATE statement
    """

    def __init__(self, read: SelectQueries, dataset: Dataset, default_named_graph: URIRef | None = None) -> None:
        super().__init__(dataset, default_named_graph)
        self._read = read

    def drop_types(
        self,
        type_: list[URIRef],
        named_graph: URIRef | None = None,
    ) -> dict[URIRef, int]:
        """Drop types from the graph store

        Args:
            type_: List of types to drop
            named_graph: Named graph to query over, default None (default graph

        Returns:
            Dictionary of dropped types
        """
        dropped_types: dict[URIRef, int] = {}
        for t in type_:
            instance_ids = list(self._read.list_instances_ids(t))
            dropped_types[t] = len(instance_ids)
            remove_instance_ids_in_batch(self.graph(named_graph), instance_ids)
        return dropped_types



from rdflib import Dataset, URIRef

from ._select import SelectQueries
from ._update import UpdateQueries


class Queries:
    """Helper class for storing standard queries for the graph store."""

    def __init__(
        self,
        dataset: Dataset,
        default_named_graph: URIRef | None = None,
    ) -> None:
        self.select = SelectQueries(dataset, default_named_graph)
        self.update = UpdateQueries(self.select, dataset, default_named_graph)



from rdflib import Dataset, Graph, URIRef
from rdflib.graph import DATASET_DEFAULT_GRAPH_ID


class BaseQuery:
    def __init__(
        self,
        dataset: Dataset,
        default_named_graph: URIRef | None = None,
    ):
        self.dataset = dataset
        self.default_named_graph = default_named_graph or DATASET_DEFAULT_GRAPH_ID

    def graph(self, named_graph: URIRef | None = None) -> Graph:
        """Get named graph from the dataset to query over"""
        return self.dataset.graph(named_graph or self.default_named_graph)



import typing
import urllib.parse
from collections.abc import Callable, Iterable, Iterator, Set
from functools import cached_property

from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling import DataModelIdentifier
from cognite.client.data_classes.data_modeling.instances import Edge, Instance, Node
from cognite.client.utils.useful_types import SequenceNotStr
from rdflib import RDF, Literal, Namespace, URIRef

from cognite.neat._client import NeatClient
from cognite.neat._config import GLOBAL_CONFIG
from cognite.neat._constants import DEFAULT_SPACE_URI, is_readonly_property
from cognite.neat._issues.errors import ResourceRetrievalError
from cognite.neat._shared import Triple
from cognite.neat._utils.collection_ import iterate_progress_bar

from ._base import BaseExtractor
from ._dict import DEFAULT_EMPTY_VALUES, DMSPropertyExtractor


class DMSExtractor(BaseExtractor):
    """Extract data from Cognite Data Fusion DMS instances into Neat.

    Args:
        total_instances_pair_by_view: A dictionary where the key is the view id and the value is a tuple with the total
            number of instances and an iterable of instances.
        limit: The maximum number of items to extract.
        overwrite_namespace: If provided, this will overwrite the space of the extracted items.
        unpack_json: If True, JSON objects will be unpacked into RDF literals.
        empty_values: If unpack_json is True, when unpacking JSON objects, if a key has a value in this set, it will be
            considered as an empty value and skipped.
        str_to_ideal_type: If unpack_json is True, when unpacking JSON objects, if the value is a string, the extractor
            will try to convert it to the ideal type.
        node_type: The prioritized order of the node type to use. The options are "view" and "type". "view"
            means the externalId of the view used as type, while type is the node.type.
        edge_type: The prioritized order of the edge type to use. The options are "view" and "type". "view"
            means the externalId of the view used as type, while type is the edge.type.
    """

    def __init__(
        self,
        total_instances_pair_by_view: dict[dm.ViewId, tuple[int | None, Iterable[Instance]]],
        limit: int | None = None,
        overwrite_namespace: Namespace | None = None,
        unpack_json: bool = False,
        empty_values: Set[str] = DEFAULT_EMPTY_VALUES,
        str_to_ideal_type: bool = False,
        node_type: tuple[typing.Literal["view", "type"], ...] = ("view",),
        edge_type: tuple[typing.Literal["view", "type"], ...] = ("view", "type"),
    ) -> None:
        self.total_instances_pair_by_view = total_instances_pair_by_view
        self.limit = limit
        self.overwrite_namespace = overwrite_namespace
        self.unpack_json = unpack_json
        self.empty_values = empty_values
        self.str_to_ideal_type = str_to_ideal_type
        self.node_type = node_type
        self.edge_type = edge_type

    @classmethod
    def from_data_model(
        cls,
        client: NeatClient,
        data_model: DataModelIdentifier,
        limit: int | None = None,
        overwrite_namespace: Namespace | None = None,
        instance_space: str | SequenceNotStr[str] | None = None,
        unpack_json: bool = False,
        str_to_ideal_type: bool = False,
    ) -> "DMSExtractor":
        """Create an extractor from a data model.

        Args:
            client: The Cognite client to use.
            data_model: The data model to extract.
            limit: The maximum number of instances to extract.
            overwrite_namespace: If provided, this will overwrite the space of the extracted items.
            instance_space: The space to extract instances from.
            unpack_json: If True, JSON objects will be unpacked into RDF literals.
        """
        retrieved = client.data_modeling.data_models.retrieve(data_model, inline_views=True)
        if not retrieved:
            raise ResourceRetrievalError(dm.DataModelId.load(data_model), "data model", "Data Model is missing in CDF")
        return cls.from_views(
            client,
            retrieved.latest_version().views,
            limit,
            overwrite_namespace,
            instance_space,
            unpack_json,
            str_to_ideal_type,
        )

    @classmethod
    def from_views(
        cls,
        client: NeatClient,
        views: Iterable[dm.View],
        limit: int | None = None,
        overwrite_namespace: Namespace | None = None,
        instance_space: str | SequenceNotStr[str] | None = None,
        unpack_json: bool = False,
        str_to_ideal_type: bool = False,
    ) -> "DMSExtractor":
        """Create an extractor from a set of views.

        Args:
            client: The Cognite client to use.
            views: The views to extract.
            limit: The maximum number of instances to extract.
            overwrite_namespace: If provided, this will overwrite the space of the extracted items.
            instance_space: The space to extract instances from.
            unpack_json: If True, JSON objects will be unpacked into RDF literals.
            str_to_ideal_type: If True, when unpacking JSON objects, if the value is a string, the extractor will try to
                convert it to the ideal type.
        """
        total_instances_pair_by_view: dict[dm.ViewId, tuple[int | None, Iterable[Instance]]] = {}
        for view in views:
            instance_iterator = _ViewInstanceIterator(client, view, instance_space)
            total_instances_pair_by_view[view.as_id()] = (instance_iterator.count, instance_iterator)

        return cls(
            total_instances_pair_by_view=total_instances_pair_by_view,
            limit=limit,
            overwrite_namespace=overwrite_namespace,
            unpack_json=unpack_json,
            str_to_ideal_type=str_to_ideal_type,
        )

    def extract(self) -> Iterable[Triple]:
        total_instances = sum(total for total, _ in self.total_instances_pair_by_view.values() if total is not None)
        use_progress_bar = (
            GLOBAL_CONFIG.use_iterate_bar_threshold and total_instances > GLOBAL_CONFIG.use_iterate_bar_threshold
        )

        for view_id, (total, instances) in self.total_instances_pair_by_view.items():
            if total == 0:
                continue
            if use_progress_bar and total is not None:
                instances = iterate_progress_bar(
                    instances,
                    total,
                    f"Extracting instances from {view_id.space}:{view_id.external_id}(version={view_id.version})",
                )

            for count, item in enumerate(instances, 1):
                if self.limit and count > self.limit:
                    break
                yield from self._extract_instance(item)

    def _extract_instance(self, instance: Instance) -> Iterable[Triple]:
        if isinstance(instance, dm.Edge):
            if not instance.properties:
                yield (
                    self._as_uri_ref(instance.start_node),
                    self._as_uri_ref(instance.type),
                    self._as_uri_ref(instance.end_node),
                )
                return
            else:
                # If the edge has properties, we create a node for the edge and connect it to the start and end nodes.
                id_ = self._as_uri_ref(instance)
                type_ = self._create_type(
                    instance, fallback=self._get_namespace(instance.space).Edge, type_priority=self.edge_type
                )
                yield id_, RDF.type, type_
                yield (
                    id_,
                    self._as_uri_ref(dm.DirectRelationReference(instance.space, "startNode")),
                    self._as_uri_ref(instance.start_node),
                )
                yield (
                    id_,
                    self._as_uri_ref(dm.DirectRelationReference(instance.space, "endNode")),
                    self._as_uri_ref(instance.end_node),
                )

        elif isinstance(instance, dm.Node):
            id_ = self._as_uri_ref(instance)
            type_ = self._create_type(
                instance, fallback=self._get_namespace(instance.space).Node, type_priority=self.node_type
            )
            yield id_, RDF.type, type_
        else:
            raise NotImplementedError(f"Unknown instance type {type(instance)}")

        if self.overwrite_namespace:
            # If the namespace is overwritten, keep the original space as a property to avoid losing information.
            yield id_, self._get_namespace(instance.space)["space"], Literal(instance.space)

        for view_id, properties in instance.properties.items():
            namespace = self._get_namespace(view_id.space)
            yield from DMSPropertyExtractor(
                id_,
                properties,
                namespace,
                self._as_uri_ref,
                self.empty_values,
                self.str_to_ideal_type,
                self.unpack_json,
            ).extract()

    def _create_type(
        self, instance: Node | Edge, fallback: URIRef, type_priority: tuple[typing.Literal["view", "type"], ...]
    ) -> URIRef:
        method_by_name: dict[str, Callable[[Node | Edge], URIRef | None]] = {
            "view": self._view_to_rdf_type,
            "type": self._instance_type_to_rdf,
        }
        for method_name in type_priority:
            type_ = method_by_name[method_name](instance)
            if type_:
                return type_
        else:
            return fallback

    def _instance_type_to_rdf(self, instance: Node | Edge) -> URIRef | None:
        if instance.type:
            return self._as_uri_ref(instance.type)
        return None

    def _view_to_rdf_type(self, instance: Node | Edge) -> URIRef | None:
        view_id = next(iter((instance.properties or {}).keys()), None)
        if view_id:
            return self._get_namespace(view_id.space)[urllib.parse.quote(view_id.external_id)]
        return None

    def _as_uri_ref(self, instance: Instance | dm.DirectRelationReference) -> URIRef:
        return self._get_namespace(instance.space)[urllib.parse.quote(instance.external_id)]

    def _get_namespace(self, space: str) -> Namespace:
        if self.overwrite_namespace:
            return self.overwrite_namespace
        return Namespace(DEFAULT_SPACE_URI.format(space=urllib.parse.quote(space)))


class _ViewInstanceIterator(Iterable[Instance]):
    def __init__(self, client: NeatClient, view: dm.View, instance_space: str | SequenceNotStr[str] | None = None):
        self.client = client
        self.view = view
        self.instance_space = instance_space

    @cached_property
    def count(self) -> int:
        node_count = edge_count = 0
        if self.view.used_for in ("node", "all"):
            node_result = self.client.data_modeling.instances.aggregate(
                view=self.view.as_id(),
                aggregates=dm.aggregations.Count("externalId"),
                instance_type="node",
                space=self.instance_space,
            ).value
            if node_result:
                node_count = int(node_result)
        if self.view.used_for in ("edge", "all"):
            edge_result = self.client.data_modeling.instances.aggregate(
                view=self.view.as_id(),
                aggregates=dm.aggregations.Count("externalId"),
                instance_type="edge",
                space=self.instance_space,
            ).value
            if edge_result:
                edge_count = int(edge_result)
        return node_count + edge_count

    def __iter__(self) -> Iterator[Instance]:
        view_id = self.view.as_id()
        read_only_properties = {
            prop_id
            for prop_id, prop in self.view.properties.items()
            if isinstance(prop, dm.MappedProperty)
            and is_readonly_property(prop.container, prop.container_property_identifier)
        }
        # All nodes and edges with properties
        if self.view.used_for in ("node", "all"):
            node_iterable: Iterable[Instance] = self.client.instances.iterate(
                instance_type="node",
                source=view_id,
                space=self.instance_space,
            )
            if read_only_properties:
                node_iterable = self._remove_read_only_properties(node_iterable, read_only_properties, view_id)
            yield from node_iterable

        if self.view.used_for in ("edge", "all"):
            yield from self.client.instances.iterate(
                instance_type="edge",
                source=view_id,
                space=self.instance_space,
            )

        for prop in self.view.properties.values():
            if isinstance(prop, dm.EdgeConnection):
                if prop.edge_source:
                    # All edges with properties are extracted from the edge source
                    continue
                yield from self.client.instances.iterate(
                    instance_type="edge",
                    filter_=dm.filters.Equals(
                        ["edge", "type"], {"space": prop.type.space, "externalId": prop.type.external_id}
                    ),
                    space=self.instance_space,
                )

    @staticmethod
    def _remove_read_only_properties(
        nodes: Iterable[Instance], read_only_properties: Set[str], view_id: dm.ViewId
    ) -> Iterable[Instance]:
        for node in nodes:
            if properties := node.properties.get(view_id):
                for read_only in read_only_properties:
                    properties.pop(read_only, None)
            yield node
